-- Big TODOlist
-- TODO: Channel:is_subscribed could be replaced by get_pep_service(...):get_subscription

local host = module:get_host();
if module:get_host_type() ~= "component" then
    error("MIX should be loaded as a component", 0);
end

local st = require("util.stanza");
local jid = require("util.jid");
local uuid = require("util.uuid");
local id = require("util.id");
local datetime = require("util.datetime");
local time = require("util.time");
local serialization = require("util.serialization");
local dataforms = require("util.dataforms");
local pep = module:depends("pep");
local helpers = module:require("mix/helpers");

local mixlib = module:require("mix/mix");
Channel = mixlib.Channel;
Participant = mixlib.Participant;

-- XML namespaces
local mix_core_xmlns = "urn:xmpp:mix:core:1";
local mix_anon_xmlns = "urn:xmpp:mix:anon:0";
--local mix_admin_xmlns = "urn:xmpp:mix:admin:0";
local mix_node_messages = "urn:xmpp:mix:nodes:messages";
local mix_node_participants = "urn:xmpp:mix:nodes:participants";
local mix_node_info = "urn:xmpp:mix:nodes:info";
--local mix_node_allowed = "urn:xmpp:mix:nodes:allowed";
--local mix_node_banned = "urn:xmpp:mix:nodes:banned";
--local mix_node_config = "urn:xmpp:mix:nodes:config";

local mam_xmlns = "urn:xmpp:mam:2";

-- Persistent data
local persistent_channels = module:open_store("mix_channels", "keyval");
local persistent_channel_data = module:open_store("mix_data", "keyval");
local message_archive = module:open_store("mix_log", "archive");

-- Configuration
local default_channel_description = module:get_option("default_description", "A MIX channel for chatting");
local default_channel_name = module:get_option("default_name", "MIX channel");
local restrict_channel_creation = module:get_option("restrict_local_channels", "local");

-- Dataforms
-- MAM
local mam_query_form = dataforms.new({
    { name = "FORM_TYPE", type = "hidden", value = mam_xmlns },
    { name = "with", type = "jid-single" },
    { name = "start", type = "text-single" },
    { name = "end", type = "text-single" },
});
-- MIX Anon
local mix_anon_form = dataforms.new({
    { name = "FORM_TYPE", type = "hidden", value = mix_anon_xmlns },
    { name = "JID Visibility", type = "text-single" },
    { name = "Private Messages", type = "text-single" },
    { name = "Presence", type = "text-single" },
    { name = "vCard", type = "text-single" },
});


-- MIX-ADMIN stuff
--[[
local mix_config_form = dataforms.new({
    { name = "FORM_TYPE", type = "hidden", value = mix_admin_xmlns },
    { name = "Last Change Made By", type = "jid-single" },
    { name = "Owner", type = "jid-multi" },
    { name = "Administrator", type = "jid-multi" },
    { name = "End of Life", type = "text-single" },
    { name = "Nodes Present", type = "list-multi" },
    { name = "Message Node Subscription", type = "list-single" },
    { name = "Presence Node Subscription", type = "list-single" },
    { name = "Participants Node Subscription", type = "list-single" },
    { name = "Information Node Subscription", type = "list-single" },
    { name = "Allowed Node Subscription", type = "list-single" },
    { name = "Banned Node Subscription", type = "list-single" },
    { name = "Configuration Node Access", type = "list-single" },
});
]]--

module:depends("disco");
module:add_identity("conference", "mix", module:get_option("name", "Prosody MIX service"));
module:add_feature("http://jabber.org/protocol/disco#info");
module:add_feature(mix_core_xmlns);

local channels = {};

local function get_channel(channel_jid)
    -- Return the channel object from the channels array for which the
    -- JID matches. If none is found, returns -1, nil
    return helpers.find(channels, function(c) return c.jid == channel_jid; end);
end

local function save_channels()
    module:log("debug", "Saving channel list...");
    local channel_list = {};
    for _, channel in pairs(channels) do
        table.insert(channel_list, channel.jid);

        persistent_channel_data:set(channel.jid, channel);
    end

    persistent_channels:set("channels", channel_list);
    module:log("debug", "Saving channel list done.");
end

function Channel:save_state()
    -- Saving the entire state everything one channel seems stupid,
    -- so we just save the changed channel
    module:log("debug", "Saving state of channel %s...", self.jid);
    persistent_channel_data:set(self.jid, self);
    module:log("debug", "Saving state done.", self.jid);
end

local function publish_participant(service, channel, spid, participant)
    -- Publish a new participant on the service
    -- NOTE: This function has be to called *after* the new participant
    --       has been added to the channel.participants attay
    service:set_node_config(mix_node_participants,
                            true,
                            { ["max_items"] = #channel.participants });
    service:publish(mix_node_participants,
                    true,
                    spid,
                    st.stanza("item", { id = spid, xmlns = "http://jabber.org/protocol/pubsub" })
                        :tag("participant", { xmlns = mix_core_xmlns })
                            :tag("nick"):text(participant["nick"]):up()
                            :tag("jid"):text(participant["jid"]));
end

function module.load()
    module:log("info", "Loading MIX channels...");

    local channel_list = persistent_channels:get("channels");
    if channel_list then
        for _, channel_data in pairs(channel_list) do
            local channel = Channel:from(persistent_channel_data:get(channel_data));
            table.insert(channels, channel);
            module:log("debug", "MIX channel %s loaded", channel.jid);
        end
    else
        module:log("debug", "No MIX channels found.");
    end
    module:log("info", "Loading MIX channels done.");
end

local function channel_not_found(stanza)
    -- Wrapper for returning a "Channel-not-found" error stanza
    return st.error_reply(stanza,
                          "cancel",
                          "item-not-found",
                          "The MIX channel was not found");
end

module:hook("host-disco-items", function(event)
    module:log("debug", "host-disco-items called");
    local reply = event.reply;
    for _, channel in pairs(channels) do
        -- Adhoc channels are supposed to be invisible
        if not channel.adhoc then
            reply:tag("item", { jid = channel.jid }):up();
        end
    end
end);

module:hook("iq/bare/http://jabber.org/protocol/disco#items:query", function(event)
    module:log("debug", "IQ-GET disco#items");

    local origin, stanza = event.origin, event.stanza;
    if stanza:get_child("query", "http://jabber.org/protocol/disco#items").attr.node ~= "mix" then
        origin.send(st.error_reply(stanza, "modify", "bad-request"));
        return true;
    end

    -- TODO: Maybe here we should check if the user has permissions to get infos
    -- about the channel before saying that it doesn't exist to prevent creating
    -- an oracle.
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    if not channel:is_participant(jid.bare(stanza.attr.from)) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end

    local reply = st.reply(stanza):tag("query", { xmlns = "http://jabber.org/protocol/disco#items", node = "mix" });
    for _, node in pairs({mix_node_messages, mix_node_participants, mix_node_info}) do
        reply:tag("item", { jid = channel.jid, node = node }):up();
    end

    origin.send(reply);
    return true;
end);

local function can_create_channels(user)
    -- Returns true when the jid is allowed to create MIX channels. False otherwise.
    -- NOTE: Taken from plugins/muc/mod_muc.lua
    local host_suffix = host:gsub("^[^%.]+%.", "");

    if restrict_channel_creation == "local" then
        module:log("debug", "Comparing %s (Sender) to %s (Host)", jid.host(user), host_suffix);

        if jid.host(user) == host_suffix then
            return true;
        else
            return false;
        end
    elseif type(restrict_channel_creation) == "table" then
        if helpers.find_str(restrict_channel_creation, user) ~= -1 then
            -- User was specifically listed
            return true;
        elseif helpers.find_str(restrict_channel_creation, jid.host(user)) then
            -- User's host was allowed
            return true;
        end

        return false;
    end

    -- TODO: Handle also true/"admin" (See mod_muc)
    return true;
end

module:hook("iq-get/host/http://jabber.org/protocol/disco#info:query", function(event)
    module:log("debug", "IQ-GET host disco#info");

    local origin, stanza = event.origin, event.stanza;
    local reply = st.reply(stanza)
                    :tag("query", { xmlns = "http://jabber.org/protocol/disco#info" })
                        -- TODO: Name
                        :tag("identity", { category = "conference", type = "mix", name = "" }):up()
                        :tag("feature", { var = "http://jabber.org/protocol/disco#info" }):up()
                        :tag("feature", { var = "http://jabber.org/protocol/disco#items" }):up()
                        :tag("feature", { var = mix_core_xmlns }):up();

    if can_create_channels(stanza.attr.from) then
        reply:tag("feature", { var = mix_core_xmlns.."#create-channel" }):up();
    end
    origin.send(reply);
    return true;
end);

module:hook("iq-get/bare/http://jabber.org/protocol/disco#info:query", function(event)
    module:log("debug", "IQ-GET disco#info");

    local origin, stanza = event.origin, event.stanza;
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end
    local reply = st.reply(stanza):tag("query", { xmlns = "http://jabber.org/protocol/disco#info" });
    reply:tag("feature", { var = "http://jabber.org/protocol/disco#info" }):up();
    reply:tag("identity", { category = "conference", name = channel.name, type = "mix" }):up();

    reply:tag("feature", { var = mix_core_xmlns }):up();
    reply:tag("feature", { var = "urn:xmpp:mam:2" }):up();

    origin.send(reply);
    return true;
end);

module:hook("iq-set/bare/"..mam_xmlns..":query", function(event)
    local stanza, origin = event.stanza, event.origin;
    local channel_jid = stanza.attr.to;
    local j, channel = get_channel(channel_jid);
    if j == -1 then
        -- TODO: Is this correct?
        origin.send(channel_not_found(stanza));
        return true;
    end

    -- Check if the user is subscribed to the messages node
    if not channel:is_subscribed(stanza.attr.from, mix_node_messages) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end

    local query = stanza:get_child("query", mam_xmlns);
    local filter = {};
    local query_id = query.attr.queryid;
    local x = query:get_child("x", "jabber:x:data");
    if x ~= nil then
        -- TODO: Error handling
        local form, err = mam_query_form:data(x);

        -- Validate
        if (form["start"] and not form["end"]) or (not form["start"] and form["end"]) then
            origin.send(st.error_reply(stanza, "modify", "bad-request", "Invalid timestamp"));
            return true;
        end

        module:log("debug", "Got a MAM query between %s and %s", form["start"], form["end"]);
        filter = {
            start = datetime.parse(form["start"]); ["end"] = datetime.parse(form["end"])
        };
    end
    local data, err = message_archive:find(channel_jid, filter);
    if not data then
        module:log("debug", "MAM error: %s", err);

        if err == "item-not-found" then
            origin.send(st.error_reply(stanza, "modify", "item-not-found"));
        else
            origin.send(st.error_reply(stanza, "cancel", "internal-server-error"));
        end
        return true;
    end
    for message_id, item, when in data do
        local msg = st.stanza("message", { from = channel_jid, to = stanza.attr.from, type = "groupchat" })
                        :tag("result", { xmlns = mam_xmlns, queryid = query_id, id = message_id })
                            :tag("forwarded", { xmlns = "urn:xmpp:forward:0" })
                                :tag("delay", { xmlns = "urn:xmpp:delay", stamp = datetime.datetime(when) }):up();
        msg:add_child(item);
        origin.send(msg);
    end
    return true;
end);

module:hook("iq-get/bare/"..mam_xmlns..":query", function(event)
    if event.stanza.attr.id ~= "form1" then return; end

    module:log("debug", "Got a MAM query for supported fields");

    local ret = st.reply(event.stanza)
                    :tag("query", { xmlns = mam_xmlns })
                        :tag("x", { xmlns = "jabber:x:data", type = "form"})
                            :tag("field", { type = "hidden", var = "FORM_TYPE" })
                                :tag("value"):text(mam_xmlns):up():up()
                            :tag("field", { type = "jid-single", var = "with" }):up()
                            :tag("field", { type = "text-single", var = "start" }):up()
                            :tag("field", { type = "text-single", var = "end" });
    module:send(ret);
    return true;
end);

module:hook("iq-set/bare/"..mix_core_xmlns..":leave", function(event)
    module:log("debug", "MIX leave received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    local j, _ = channel:find_participant(from);
    if j == -1 then
        origin.send(st.error_reply(stanza,
                                   "cancel",
                                   "forbidden",
                                   "Not a participant"));
        channel:debug_print();
        module:log("debug", "%s is not a participant in %s", from, channel.jid);
        return true;
    end

    -- Remove the user as a participant by...
    -- Unsubscribing
    local srv = pep.get_pep_service(jid.node(channel.jid));
    for _, node in pairs(channel.subscriptions[from]) do
        srv:set_affiliation(node, true, from, "outcast");
        srv:remove_subscription(node, true, from);
        module:log("debug", "Unsubscribed %s from %s on %s", from, node, channel.jid);
    end
    channel.subscriptions[from] = nil;
    -- Retracting the participation
    local spid = channel:get_spid(from);
    local notifier = st.stanza("retract", { id = spid });
    -- TODO: Maybe error handling
    srv:retract(mix_node_participants, true, spid, notifier);
    -- Removing the user
    table.remove(channel.participants, j);
    channel:save_state();

    origin.send(st.reply(stanza):tag("leave", { xmlns = mix_core_xmlns }));
    return true;
end);

module:hook("iq-set/bare/"..mix_core_xmlns..":join", function(event)
    module:log("debug", "MIX join received");

    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    -- Prevent the user from joining multiple times
    local j, _ = channel:find_participant(from);
    if j ~= -1 then
        module:send(st.error_reply(stanza, "cancel", "bad-request", "User already joined"));
        return true;
    end

    local spid = channel:get_spid(from) or uuid.generate(); -- Stable Participant ID
    local reply = st.reply(stanza)
                    :tag("join", { xmlns = mix_core_xmlns, id = spid });
    local srv = pep.get_pep_service(jid.node(stanza.attr.to));
    local join = stanza:get_child("join", mix_core_xmlns);
    local nick_tag = join:get_child("nick");
    local nick;
    if not nick_tag then
        nick = jid.node(from);
    else
        nick = nick_tag:get_text();
    end
    module:log("debug", "User joining as nick %s", nick);

    local nodes = {};
    local has_subscribed_once = false;
    local first_error = nil;
    for subscribe in join:childtags("subscribe") do
        module:log("debug", "Subscribing user to node %s", subscribe.attr.node);
        -- TODO: Once MIX-ADMIN is implemented, we should check here what
        -- affiliation we set, e.g. if the JID is the owner, then set owner.
        srv:set_affiliation(subscribe.attr.node, true, from, "member");
        local ok, err = srv:add_subscription(subscribe.attr.node, true, from);
        if not ok then
            module:log("debug", "Error during subscription: %s", err);

            -- MIX-CORE says that the first error should be returned when
            -- no of the requested nodes could be subscribed to
            if first_error ~= nil then
                first_error = err;
            end
        else
            table.insert(nodes, subscribe.attr.node);
            reply:tag("subscribe", { node = subscribe.attr.node }):up();
            has_subscribed_once = true;
        end
    end

    if not has_subscribed_once then
        origin.send(st.error_reply(stanza, "cancel", first_error));
        return true;
    end

    -- TODO: Make the default configurable
    local jid_visibility = "never"; -- default
    local allow_pms = "allow";
    local allow_vcards = "block";
    local share_presence = "share";
    local x = join:get_child("x", "jabber:x:data");
    if x ~= nil then
        -- TODO: Rethink naming
        -- TODO: Error handling?
        local form, err = mix_anon_form:data(x);
        if form["JID Visibility"] then
            jid_visibility = form["JID Visibility"];
        end
        if form["Private Messages"] then
            allow_pms = form["Private Messages"];
        end
        if form["Presence"] then
            share_presence = form["Presence"];
        end
        if form["vCard"] then
            allow_vcards = form["vCard"];
        end
    end

    local participant = Participant:new(jid.bare(from), nick, jid_visibility, allow_pms, share_presence, allow_vcards);
    channel.subscriptions[from] = nodes;
    table.insert(channel.participants, participant)
    channel:set_spid(jid.bare(stanza.attr.from), spid);
    publish_participant(srv, channel, spid, participant);
    channel:save_state();

    reply:add_child(nick);
    reply:tag("x", { xmlns = "jabber:x:data", type = "result" })
        :tag("field", { var = "FORM_TYPE", type = "hidden" })
            :tag("value"):text(mix_anon_xmlns):up():up()
        :tag("field", { var = "JID Visibility"})
            :tag("value"):text(jid_visibility):up():up()
        :tag("field", { var = "Private Messages"})
            :tag("value"):text(allow_pms):up():up()
        :tag("field", { var = "Presence"})
            :tag("value"):text(share_presence):up():up()
        :tag("field", { var = "vCard"})
            :tag("value"):text(allow_vcards):up():up();

    origin.send(reply);
    return true
end);

module:hook("iq-set/bare/"..mix_core_xmlns..":setnick", function(event)
    module:log("debug", "MIX setnick received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    local j, participant = channel:find_participant(from);
    if j == -1 then
        channel:debug_print();
        module:log("debug", "%s is not a participant in %s", from, channel.jid);
        return true;
    end

    -- NOTE: Prosody should guarantee us that the setnick stanza exists
    local setnick = stanza:get_child("setnick", mix_core_xmlns);
    local nick = setnick:get_child("nick");
    if nick_stanza == nil then
        origin.send(st.error_reply(stanza, "cancel", "bad-request", "Missing <nick>"));
        return true;
    end

    -- Change the nick
    channel.participants[j].nick = nick:get_text();
    -- Inform all other members
    local srv = pep.get_pep_service(channel.jid);
    --local participant = channel.participants[participant_index];
    publish_participant(srv, channel, channel:get_spid(participant.jid), participant);

    origin.send(st.reply(stanza)
                    :tag("setnick", { xmlns = mix_core_xmlns })
                        :tag("nick"):text(nick:get_text()));
    channel:save_state();
    return true;
end);

function Channel:publish_info(srv)
    local timestamp = datetime.datetime(time.now());
    local info = st.stanza("item", { id = timestamp, xmlns = "http://jabber.org/protocol/pubsub" })
                    :tag("x", { xmlns = "jabber:x:data", type = "result" })
                        :tag("field", { var = "FORM_TYPE", type = "hidden" })
                            :tag("value"):text(mix_core_xmlns):up():up()
                        :tag("field", { var = "Name" })
                            :tag("value"):text(self.name):up()
                        :tag("field", { var = "Contact" });
    for _, contact in pairs(self.contacts) do
        info:add_child(st.stanza("value"):text(contact));
    end
    srv:publish(mix_node_info, true, timestamp, info);
end

local function create_channel(node, creator, adhoc)
    local channel = Channel:new(string.format("%s@%s", node, host),
                                default_channel_name,
                                default_channel_description,
                                {},
                                {},
                                {},
                                { creator },
                                adhoc);
    -- Create the PEP nodes
    local srv = pep.get_pep_service(node);
    -- MIX-CORE
    for _, psnode in pairs({ mix_node_info, mix_node_participants, mix_node_messages }) do
        srv:create(psnode, true, {
            ["access_model"] = "whitelist",
            ["persist_items"] = true,
        });
    end
    channel:publish_info(srv);

    --[[
    -- MIX-ADMIN
    local admin_nodes = { mix_node_banned, mix_node_config };
    if adhoc then
        table.insert(admin_nodes, mix_node_allowed);
    end
    for _, psnode in pairs(admin_nodes) do
        srv:create(mix_node_allowed, true, { ["access_model"] = "whitelist" });
        srv:set_affiliation(mix_node_allowed, true, creator, "owner");
    end
    if adhoc then
        -- Allow the creator to join
        srv:publish(mix_node_allowed,
                    true,
                    nil,
                    st.stanza("item", { id = creator }));
    end
    ]]--
    table.insert(channels, channel);
end

module:hook("iq-set/host/"..mix_core_xmlns..":create", function(event)
    module:log("debug", "MIX create received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);

    -- Check permissions
    if not can_create_channels(from) then
        origin.send(st.error_reply(stanza, "cancel", "forbidden", "Not authorized to create channels"));
        return true;
    end

    local create = stanza:get_child("create", mix_core_xmlns);
    local node;
    if create.attr.channel ~= nil then
        -- Create non-adhoc channel
        module:log("debug", "Attempting to create channel %s", create.attr.channel);
        node = create.attr.channel;
        local i, _ = get_channel(create.attr.channel.."@"..stanza.attr.to);
        module:log("debug", "Channel index %s", i);
        if i ~= -1 then
            origin.send(st.error_reply(stanza,
                                       "cancel",
                                       "conflict",
                                       "Channel already exists"));
            return true;
        end

        create_channel(create.attr.channel, from, false);
    else
        -- Create adhoc channel
        while (true) do
            node = id.short();
            local i, _ = get_channel(string.format("%s@%s", node, host));
            if i == -1 then
                break;
            end
        end

        create_channel(node, from, true);
    end
    module:log("debug", "Channel %s created with %s as owner", node, from);

    origin.send(st.reply(stanza)
                :tag("create", { xmlns = mix_core_xmlns, channel = node }));
    save_channels();
    return true;
end);

module:hook("iq-set/host/"..mix_core_xmlns..":destroy", function(event)
    module:log("debug", "MIX destroy received");
    local origin, stanza = event.origin, event.stanza;
    local from = jid.bare(stanza.attr.from);

    local destroy = stanza:get_child("destroy", mix_core_xmlns);
    local node = destroy.attr.channel;
    local node_jid = string.format("%s@%s", node, host);
    local i, channel = get_channel(node_jid);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    -- TODO: Maybe make this configurable
    if helpers.find_str(channel.contacts, from) == -1 then
        origin.send(st.error_reply(stanza, "cancel", "forbidden"));
        return true;
    end
    -- Remove all registered nodes
    local srv = pep.get_pep_service(node);
    for _, pep_node in pairs({ mix_node_participants, mix_node_info, mix_node_messages }) do
        srv:delete(pep_node, true);
    end
    table.remove(channels, i);

    module:log("debug", "Channel %s destroyed", node);

    origin.send(st.reply(stanza));
    save_channels();
    return true;
end);

module:hook("message/bare", function(event)
    module:log("debug", "MIX message received");
    local stanza, origin = event.stanza, event.origin;
    if stanza.attr.type ~= "groupchat" then
        origin.send(st.error_reply(stanza, "modify", "bad-request", "Non-groupchat message"));
        return true;
    end

    local from = jid.bare(stanza.attr.from);
    local _, channel = get_channel(stanza.attr.to);
    if not channel then
        origin.send(channel_not_found(stanza));
        return true;
    end

    local j, participant = channel:find_participant(from);
    if j == -1 then
        origin.send(st.error_reply(stanza, "cancel", "forbidden", "Not a participant"));
        return true;
    end

    local msg = st.clone(stanza);
    msg:add_child(st.stanza("mix", { xmlns = mix_core_xmlns })
                    :tag("nick"):text(participant.nick):up()
                    :tag("jid"):text(participant.jid):up());

    -- Put the message into the archive
    local mam_id = uuid.generate();
    msg.attr.id = mam_id;
    -- NOTE: The spec says to do so
    msg.attr.from = channel.jid;
    message_archive:append(stanza.attr.to, mam_id, msg, time.now());
    msg.attr.from = channel.jid.."/"..channel:get_spid(from);

    if module:fire_event("mix-broadcast-message", msg) then
        return true;
    end

    for _, p in pairs(channel.participants) do
        -- Only users who subscribed to the messages node should receive
        -- messages
        if channel:is_subscribed(p.jid, mix_node_messages) then
            local tmp = st.clone(msg);
            tmp.attr.to = p.jid;
            module:send(tmp);
        end
    end
    return true;
end);
